from focal_loss.focal_loss import FocalLoss
from .vistabnet import TabularVisionTransformer
import numpy as np
import torch
import tqdm


def balanced_accuracy(y_true, y_pred):
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)

    classes = np.unique(y_true)
    accs = []
    for c in classes:
        idx = y_true == c
        acc = (y_true[idx] == y_pred[idx]).sum() / idx.sum()
        accs.append(acc)
    return np.mean(accs)

class VisTabNetClassifier:
    def __init__(self, input_features, classes, projections=195, projection_depth=4, lr=1e-4, proj_lr=3e-4, device="cuda:0"):
        self.device = device
        self.device_type = torch.device(device).type

        self.input_features = input_features
        self.classes = classes
        self.projections = projections
        self.projection_depth = projection_depth

        model = TabularVisionTransformer(self.input_features, self.classes, self.projections, self.projection_depth,
            image_size=224,
            patch_size=16,
            num_layers=12,
            num_heads=12,
            hidden_dim=768,
            mlp_dim=3072
        ).to(torch.float16).to(self.device)

        self.model = model

        # Prepare optimizer
        params = list(filter(lambda p: p.requires_grad, self.model.head.parameters()))
        params_projector = list(filter(lambda p: p.requires_grad, self.model.projector.parameters()))
        self.optim = torch.optim.Adam([
        {
            'params': params,
            'lr': lr,
        },
        {
            'params': params_projector,
            'lr': proj_lr,
        }
        ], eps=1e-4)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optim, lambda epoch: 0.98 ** epoch)
        self.criterion = FocalLoss(gamma=1.5)

    def _prepare_dataloader(self, X, y, batch_size, shuffle=False):
        X = torch.tensor(X)
        y = torch.tensor(y)
    
        dataset = torch.utils.data.TensorDataset(X, y)
        loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)
        return loader

    def fit(self, X, y, eval_X=None, eval_y=None, lr=1e-4, proj_lr=3e-4, epochs=100, batch_size=32, verbose=True):
        NOISE_STD = 0.01

        def training_step(model, optim, criterion, X, y):
            X = X.to(self.device).to(torch.float16)
            y = y.to(self.device)

            X = X + torch.randn(X.shape).to(self.device) * NOISE_STD

            optim.zero_grad()
            with torch.autocast(device_type=self.device_type, dtype=torch.float16, enabled=True):
                y_pred_ = model(X)

            assert torch.all(torch.isfinite(y_pred_))
            assert torch.all(y_pred_ >= 0)
            assert torch.all(y_pred_ <= 1)
            loss = criterion(y_pred_, y)
            loss.backward()
            optim.step()
            return y,y_pred_, loss.item()

        # Prepare dataloaders
        trainloader = self._prepare_dataloader(X, y, batch_size=batch_size, shuffle=True)
        testloader = None
        if eval_X is not None and eval_y is not None:
            testloader = self._prepare_dataloader(eval_X, eval_y, batch_size=batch_size, shuffle=False)

        # Train
        with tqdm.trange(epochs, disable=not verbose) as t:
            for epoch in t:
                self.model.train()
                epoch_loss = 0
                y_true = []
                y_pred = []
                for X, y in trainloader:
                    y, y_pred_, ls = training_step(self.model, self.optim, self.criterion, X, y)
                    epoch_loss += ls
                    
                    y_true.extend(y.cpu().tolist())
                    y_pred.extend(y_pred_.argmax(dim=1).cpu().tolist())

                train_bacc = balanced_accuracy(y_true, y_pred)
                self.scheduler.step()
                self.model.eval()

                if testloader is not None:
                    y_true = []
                    y_pred = []
                    with torch.inference_mode():
                        for X, y in testloader:
                            X = X.to(self.device).to(torch.float16)
                            y = y.to(self.device)
                            with torch.autocast(device_type=self.device_type, dtype=torch.float16, enabled=True):
                                y_pred_ = self.model(X)

                            y_true.extend(y.cpu().tolist())
                            y_pred.extend(y_pred_.argmax(dim=1).cpu().tolist())

                    bacc = balanced_accuracy(y_true, y_pred)
                    t.set_description(f"Epoch {epoch} loss: {epoch_loss:.4f} train_bacc: {train_bacc:.4f} eval_bacc: {bacc:.4f}")
                else:
                    t.set_description(f"Epoch {epoch} loss: {epoch_loss:.4f} train_bacc: {train_bacc:.4f}")
                    
            
    def predict(self, X):
        batch_size = 32
        self.model.eval()
        with torch.inference_mode():
            with torch.autocast(device_type=self.device_type, dtype=torch.float16, enabled=True):
                y_pred = []
                for i in range(0, len(X), batch_size):
                    X_ = X[i:i+batch_size]
                    X_ = torch.tensor(X_).to(self.device).to(torch.float16)
                    y_pred_ = self.model(X_)
                    y_pred.append(y_pred_.cpu())
                y_pred = torch.cat(y_pred, dim=0)
            y_pred = y_pred.argmax(dim=1).tolist()
        return y_pred

    
    def predict_proba(self, X):
        batch_size = 32

        self.model.eval()
        with torch.inference_mode():
            with torch.autocast(device_type=self.device_type, dtype=torch.float16, enabled=True):
                y_pred = []
                for i in range(0, len(X), batch_size):
                    X_ = X[i:i+batch_size]
                    X_ = torch.tensor(X_).to(self.device).to(torch.float16)
                    y_pred_ = self.model(X_).cpu()
                    y_pred.append(y_pred_)
                y_pred = torch.cat(y_pred, dim=0)
        return y_pred
  


import torch
from torchvision.ops.misc import MLP
from torch import nn
from functools import partial
from typing import Callable
from collections import OrderedDict

class MLPBlock(MLP):
    """Transformer MLP block."""

    _version = 2

    def __init__(self, in_dim: int, mlp_dim: int, dropout: float):
        super().__init__(in_dim, [mlp_dim, in_dim], activation_layer=nn.GELU, inplace=None, dropout=dropout)

        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.normal_(m.bias, std=1e-6)

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        version = local_metadata.get("version", None)

        if version is None or version < 2:
            # Replacing legacy MLPBlock with MLP. See https://github.com/pytorch/vision/pull/6053
            for i in range(2):
                for type in ["weight", "bias"]:
                    old_key = f"{prefix}linear_{i+1}.{type}"
                    new_key = f"{prefix}{3*i}.{type}"
                    if old_key in state_dict:
                        state_dict[new_key] = state_dict.pop(old_key)

        super()._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )
        
class EncoderBlock(nn.Module):
    """Transformer encoder block."""

    def __init__(
        self,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: int,
        dropout: float,
        attention_dropout: float,
        norm_layer: Callable[..., torch.nn.Module] = partial(nn.LayerNorm, eps=1e-6),
    ):
        super().__init__()
        self.num_heads = num_heads

        # Attention block
        self.ln_1 = norm_layer(hidden_dim)
        self.self_attention = nn.MultiheadAttention(hidden_dim, num_heads, dropout=attention_dropout, batch_first=True)
        self.dropout = nn.Dropout(dropout)

        # MLP block
        self.ln_2 = norm_layer(hidden_dim)
        self.mlp = MLPBlock(hidden_dim, mlp_dim, dropout)

    def forward(self, input: torch.Tensor, mask: torch.Tensor=None):
        torch._assert(input.dim() == 3, f"Expected (batch_size, seq_length, hidden_dim) got {input.shape}")
        x = self.ln_1(input)
        x, _ = self.self_attention(x, x, x, need_weights=False, attn_mask=mask)
        x = self.dropout(x)
        x = x + input

        y = self.ln_2(x)
        y = self.mlp(y)
        return x + y


class Encoder(nn.Module):
    """Transformer Model Encoder for sequence to sequence translation."""

    def __init__(
        self,
        seq_length: int,
        num_layers: int,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: int,
        dropout: float,
        attention_dropout: float,
        norm_layer: Callable[..., torch.nn.Module] = partial(nn.LayerNorm, eps=1e-6),
    ):
        super().__init__()
        # Note that batch_size is on the first dim because
        # we have batch_first=True in nn.MultiAttention() by default
        self.pos_embedding = nn.Parameter(torch.empty(1, seq_length, hidden_dim).normal_(std=0.02))  # from BERT
        self.dropout = nn.Dropout(dropout)
        layers: OrderedDict[str, nn.Module] = OrderedDict()
        for i in range(num_layers):
            layers[f"encoder_layer_{i}"] = EncoderBlock(
                num_heads,
                hidden_dim,
                mlp_dim,
                dropout,
                attention_dropout,
                norm_layer,
            )
        self.layers = nn.Sequential(layers)
        self.ln = norm_layer(hidden_dim)

    def forward(self, input: torch.Tensor, mask: torch.Tensor):
        torch._assert(input.dim() == 3, f"Expected (batch_size, seq_length, hidden_dim) got {input.shape}")
        input = input + self.pos_embedding
        dropped = self.dropout(input)
        first_layer = self.layers[0]
        mask = mask.repeat((mask.shape[1], 1))
        res = first_layer(dropped, mask)

        return self.ln(self.layers[1:](res))




from .vistabnet_backbone import Encoder
from torchvision.models.vision_transformer import VisionTransformer, ViT_B_16_Weights
from torch import nn
from functools import partial

import torch

from functools import partial

class TabularVisionTransformer(torch.nn.Module):
    class TabularVit(VisionTransformer):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.embed_dim = kwargs["hidden_dim"]
            self.encoder = Encoder(
                197,
                kwargs["num_layers"],
                kwargs["num_heads"],
                kwargs["hidden_dim"],
                kwargs["mlp_dim"],
                0.0,
                0.0,
                partial(nn.LayerNorm, eps=1e-6),
            )

            weights = ViT_B_16_Weights.DEFAULT.get_state_dict(progress=True)
            self.load_state_dict(weights)

        def _preprocess_input(self, x):
            return x

        def forward(self, x, mask=None):
            b, emb, f = x.shape
            x = torch.cat((self.class_token.expand(b, -1, -1), x), dim=1)
            if hasattr(self, 'positional_embedding'): 
                x = self.positional_embedding(x)

            if mask is not None:
                mask = torch.cat((torch.ones(1, 1).to(x.device), mask), dim=1).to(torch.bool)
                
            x = self.encoder(x, mask)
            if hasattr(self, 'pre_logits'):
                x = self.pre_logits(x)
                x = torch.tanh(x)
            return x
        
    class Projector(torch.nn.Module):
        # 196 projections which will be computed in parallel
        def __init__(self, input_size, projections, proj_depth, output_size):
            super().__init__()
            self.projections = projections

            self.projector = torch.nn.Parameter(torch.rand(self.projections, input_size, output_size))
            self.bias = torch.nn.Parameter(torch.rand(self.projections, output_size))

            # self.projector_hidden = torch.nn.Linear(output_size, output_size)
            self.projector_hidden = []
            for i in range(proj_depth): 
                self.projector_hidden.append(torch.nn.Linear(output_size, output_size))
                self.projector_hidden.append(torch.nn.GELU())
                self.projector_hidden.append(torch.nn.LayerNorm(output_size, eps=1e-6))

            self.projector_hidden = torch.nn.Sequential(*self.projector_hidden)

            self.activation = torch.nn.GELU()
            self.layernorm = torch.nn.LayerNorm(output_size, eps=1e-6)
            # self.batchnorm = torch.nn.BatchNorm1d(output_size)

        def forward(self, x):
            #input shape: (batch_size, input_size)
            #output shape: (batch_size, 196, output_size)
            x = x.unsqueeze(1).repeat(1, self.projections, 1)
            x = torch.einsum("bei,eio->beo", x, self.projector)
            x = x + self.bias
            x = self.activation(x)
            x = self.layernorm(x)

            x = self.projector_hidden(x)

            # x = torch.einsum("bei,eio->beo", x, self.projector_hidden)
            # x = x + self.hidden_bias
            # x = self.activation(x)
            return x
        
    class Head(torch.nn.Module):
        def __init__(self, input_size, output_size):
            super().__init__()
            self.head = torch.nn.Linear(input_size, 128)
            self.head_hidden = torch.nn.Linear(128, 128)
            self.head2 = torch.nn.Linear(128, output_size)
            self.softmax = torch.nn.Softmax(dim=1)
            self.relu = torch.nn.ReLU()

        def forward(self, x):
            x = self.head(x)
            x = self.relu(x)
            x = self.head_hidden(x)
            x = self.relu(x)
            x = self.head2(x)
            x = self.softmax(x)
            return x
        
    def __init__(self, input_size, output_size, projections, proj_depth, *args, **kwargs):
        super().__init__()
        self.vit = self.TabularVit(*args, **kwargs)
        self.projector = self.Projector(input_size, projections, proj_depth, self.vit.embed_dim)
        self.head = self.Head(self.vit.embed_dim, output_size)

        self.vit.requires_grad_(False)

    def forward(self, x):
        x = self.projector(x)
        mask = torch.ones(1, x.shape[1]).to(x.device)
        mask = torch.cat((mask, torch.zeros(1, 196 - x.shape[1]).to(x.device)), dim=1)

        padding = torch.zeros(x.shape[0], 196 - x.shape[1], x.shape[2]).to(x.device)
        x = torch.cat((x, padding), dim=1)

        x = self.vit(x, mask)[:, 0, :]
        x = self.head(x)
        return x